#!/usr/bin/python
#
# Copyright 2014 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import azuresecurity.util as util
import os
import string
import sys
import syslog
from azuresecurity.util import waagent
from defusedxml.ElementTree import parse

# Initialize waagent
waagent.LoggerInit('/var/log/waagent.log','/dev/stdout')

# General Settings
configFile = util.configDir + "/scanner.xml"
baselineScript = util.scriptDir + "/apply-secbaseline.sh"
   
# Parse command line arguments, set options as appropriate
def ParseArguments():
    global configFile
    parser = argparse.ArgumentParser(description='Enable/disable security scanning options [' + configFile + ']')
    parser.add_argument(
        '-v', '--verbose', help='Write output to stdout', action='count')
    parser.add_argument(
        '-l', '--list', help='List scans along with their status (enabled/disabled)', action='store_true')
    parser.add_argument(
        '-a', '--avscan', help='Enable or disable antivirus scan')
    parser.add_argument(
        '-b', '--baselinescan', help='Enable or disable  baseline configuration scan')
    parser.add_argument(
        '-f', '--file', help='use the specified config file')
    parser.add_argument(
        '-p', '--patchscan', help='Enable or disable  patch scan')
    parser.add_argument(
        '-s', '--softwarescan', help='Enable or disable  software inventory scan')
    parser.add_argument(
        '-ab', '--applybaseline', help='*Applies secure baseline settings to system*', action='store_true')
    args = parser.parse_args()
    # Set verbosity level for writing to console
    if args.verbose:
        util.runtimeVerbosity = args.verbose
    if args.file:
        configFile=args.file
    return args

def ParseConfig(configFile):
    if not configFile:
        util.WriteConsole(syslog.LOG_ERR, "Config file path must be a valid string")
        return False, 0
    if not os.path.isfile(configFile):
        util.WriteConsole(syslog.LOG_ERR, "No config file exists at the path" + configFile)
        return False, 0
    # Parse the file
    try:
        util.WriteVerbose("Parsing config file " + configFile)
        elementTree = parse(configFile)
        return True, elementTree
    except:
        util.WriteConsole(syslog.LOG_ERR, "Unable to parse file as XML: " +  configFile)
        return False, 0  
        
def ApplyBaseline():
    print "**WARNING - This will script will make changes to your system! **"
    print "For additional details visit: http://aka.ms/jrt4vl"
    response = raw_input("Are you sure you want to proceed?[YES/NO]")
    if not (response == "YES"):
        util.WriteConsole(syslog.LOG_INFO,"Installation of secure baseline was cancelled by user input")
        return 0
    util.WriteConsole(syslog.LOG_INFO,"Initiating installation of secure baseline")
    error,msg = waagent.RunGetOutput(baselineScript)
    lines = string.split(msg, "\n")
    for line in lines:
        util.WriteLog(syslog.LOG_INFO, line)
    logpri = syslog.LOG_WARNING if error else syslog.LOG_INFO
    util.WriteConsole(logpri,"Installation of secure baseline is complete [{0}]".format(error))
    return 0
        
# Return true if config is actually modified, false if not
def ModifyConfig(config, scan, desiredState):
    if not (scan or desiredState) or config is None:
        util.WriteConsole(syslog.LOG_ERR, "An invalid parameter was passed while attempting to modify config")
        return False
    util.WriteDebug("Processing request for " + scan)
    # What happens if elements don't exist
    scanners = config.find('scanners')
    scanState = scanners.get(scan)
    if (util.IsEnabled(desiredState)):
        desiredState = util.enabled
    elif (util.IsDisabled(desiredState)):
        desiredState = util.disabled
    else:
        util.WriteConsole(syslog.LOG_ERR, "* [{0}] Command not recognized: {1}. Please use 'enable' or 'disable'".format(scan, desiredState))
        return False
    util.WriteDebug("Request to set {0} to {1}".format(scan, desiredState))
    if util.IsConfigMatch(scanState, desiredState):
        util.WriteConsole(syslog.LOG_INFO, "* {0} is already {1}".format(scan, desiredState))
        return False
    scanners.attrib[scan] = desiredState
    util.WriteConsole(syslog.LOG_INFO, "* Set {0} to {1}".format(scan, desiredState))
    return True

# Parse arguments and config file, kick off scans, report results
def main():
    args = ParseArguments()
    success, elementTree = ParseConfig(configFile)
    if not success:
        sys.exit(1)
    modified = False
    config = elementTree.getroot()
    exitCode = 0
    if (args.applybaseline):
        exitCode += ApplyBaseline()
    if (args.avscan):
        modified |= ModifyConfig(config, "avscan", args.avscan)
    if (args.baselinescan):
        modified |= ModifyConfig(config, "baselinescan", args.baselinescan)
    if (args.patchscan):
        modified |= ModifyConfig(config, "patchscan", args.patchscan)
    if (args.softwarescan):
        modified |= ModifyConfig(config, "softwarescan", args.softwarescan)
    if (args.list):
        util.WriteDebug("Processing 'list' command")
        scanners = config.find('scanners')
        for key in scanners.keys():
            print "{0:<17s} {1}".format(key, scanners.get(key))
    if modified:
        util.WriteLog(syslog.LOG_INFO, "Writing updated config to " + configFile)
        with open(configFile, 'w') as ofile:
            elementTree.write(ofile)
    sys.exit(exitCode)

main()

