#!/usr/bin/python
#
# Copyright 2014 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import array
import hashlib
import inspect
import os
import os.path
import stat
import string
import sys
import syslog

import azuresecurity.avscan as avscan
import azuresecurity.baselinescan as baselinescan
import azuresecurity.patchscan as patchscan
import azuresecurity.softwarescan as softwarescan
import azuresecurity.util as util
from azuresecurity.util import waagent
from defusedxml.ElementTree import parse

# Initialize waagent
waagent.LoggerInit('/var/log/waagent.log','/dev/null')

# General Settings
configFile = util.configDir + "/scanner.xml"
   
# Parse command line arguments, set options as appropriate
def ParseArguments():
    parser = argparse.ArgumentParser(description='Scan the system and compare with baseline security standard')
    parser.add_argument(
        '-v', '--verbose', help='write output to stdout', action='count')
    parser.add_argument(
        '-a', '--avscan', help='perform an antivirus scan', action='store_true')
    parser.add_argument(
        '-b', '--baselinescan', help='perform a baseline configuration scan', action='store_true')
    parser.add_argument(
        '-c', '--check', help='run only the specified check (by msid)')
    parser.add_argument(
        '-d', '--default', help='perform default scans by reading config file located at ' + configFile, action='store_true')
    parser.add_argument(
        '-f', '--file', help='perform scans per the specified config file')
    parser.add_argument(
        '-p', '--patchscan', help='perform a patch scan', action='store_true')
    parser.add_argument(
        '-s', '--softwarescan', help='perform a software inventory scan', action='store_true')
    args = parser.parse_args()
    # Set verbosity level for writing to console
    if args.verbose:
        util.runtimeVerbosity = args.verbose
    if (args.default):
        args.file = configFile
    return args

def ParseConfig(args):
    if not args.file:
        util.WriteLog(syslog.LOG_ERR, "Config file path must be a valid string")
        return
    if not os.path.isfile(args.file):
        util.WriteLog(syslog.LOG_ERR, "No config file exists at the path" + args.file)
        return
    # Parse the file
    util.WriteVerbose("Parsing config file " + args.file)
    try:
        elementTree = parse(args.file)
        root = elementTree.getroot()
        scanners = root.find('scanners')
    except:
        util.WriteConsole(syslog.LOG_ERR, "Unable to parse file as XML: " +  args.file)
        return
    args.avscan = util.IsEnabled(scanners.get('avscan'))
    args.baselinescan = util.IsEnabled(scanners.get('baselinescan'))
    args.patchscan = util.IsEnabled(scanners.get('patchscan'))
    args.softwarescan = util.IsEnabled(scanners.get('softwarescan'))
    if not (args.avscan or args.baselinescan or args.patchscan or args.softwarescan):
        util.WriteConsole(syslog.LOG_ERR, "No scans enabled in " + args.file)
        sys.exit(1)
    
def ParseArgumentsAndConfig():
    args = ParseArguments()
    if (args.file):
        ParseConfig(args)
    if not (args.avscan or args.baselinescan or args.patchscan or args.softwarescan):
        print "No scans specified. Run " + __file__ + " -h for help"
        sys.exit(1)
    return args
        
# Parse arguments and config file, kick off scans, report results
def main():
    args = ParseArgumentsAndConfig()
    findings = 0
    attemptedScans = 0
    errCode = 0
    if (args.avscan):
        e, f = avscan.UpdateAndScan()
        errCode += e
        findings += f
        attemptedScans += 1
    if (args.baselinescan):
        e, f = baselinescan.RunSecureBaselineScan(args.check)
        errCode += e
        findings += f
        attemptedScans += 1
    if (args.patchscan):
        e, f = patchscan.RunPatchScan()
        errCode += e
        findings += f
        attemptedScans += 1
    if (args.softwarescan):
        e, f = softwarescan.RunSoftwareInventoryScan()
        errCode += e
        findings += f
        attemptedScans += 1
    logpri = syslog.LOG_WARNING if errCode else syslog.LOG_INFO
    util.WriteLog(logpri, "Completed " + `attemptedScans` + " scan(s) with " + `errCode` + " errors and " + `findings` + " findings")
    sys.exit(errCode)

main()
