#!/usr/bin/python
#
# Copyright 2014 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import hashlib
import inspect
import os
import os.path
import re
import stat
import string
import syslog
import util
import patchscan
from util import waagent
from defusedxml.ElementTree import parse

# Sha256 sum of the audits.xml file, to confirm it has not been modified
auditsHash = "0e9829cada19a76be91062f0a39437e9a5fb0010d818993be7447fb05f5b65a2"
# To be updated in next iteration
auditsXml = util.configDir + "/audits.xml"

# Return True if element is a valid object and contains each of the required attributes
def ValidateElement(element, *args):
    caller = "(" + inspect.stack()[1][3] + ")"
    util.WriteDebug("Validating elements for " + caller)
    try:
        if element is None:
            msg = "Xml element failed validation" + caller
            util.WriteLog(syslog.LOG_ERR, msg)
            return False, msg
        for attribute in args:
            if not element.get(attribute):
                msg = "Xml element is missing a necessary attribute: " + attribute + " " + caller
                util.WriteLog(syslog.LOG_ERR, msg)
                return False, msg
    except:
        util.WriteLog(syslog.LOG_ERR, "Exception occurred while parsing xml element")
        return False
    return True, 0
    
# Return true only if the file does exist
def CheckFileExists(checkElement):
    success, msg = ValidateElement(checkElement, "path")
    if not success:
        return False, msg
    path = checkElement.get("path")
    if (os.path.exists(path)):
        return True, 0
    return False, "File " + path + " not found"

# Return true only if the file does not exist
def CheckFileNotExists(checkElement):
    success, msg = ValidateElement(checkElement, "path")
    if not success:
        return False, msg
    path = checkElement.get("path")
    if (os.path.exists(path)):
        return False, "File " + path + " found"
    return True, 0
    
# Return true only if the package is installed
def CheckPackageInstalled(checkElement):
    success, msg = ValidateElement(checkElement, "packagename")
    if not success:
        return False, msg
    packagename = checkElement.get("packagename")
    # Check for package. Noop at the end to hide error code from grep
    command = "dpkg -l | egrep '^ii\s+{0}\s'; :".format(packagename)
    error,msg = waagent.RunGetOutput(command)
    msg = msg.strip()
    if ( len(msg) < 1 ):
        return False, "Package " + packagename + " not installed"
    return True, 0

# Return true only if the package is not installed
def CheckPackageNotInstalled(checkElement):
    success, msg = ValidateElement(checkElement, "packagename")
    if not success:
        return False, msg
    packagename = checkElement.get("packagename")
    success, msg = CheckPackageInstalled(checkElement)
    if success:
        return False, "Package " + packagename + " is installed"
    return True, 0

# Return true only if file exists and permissions match
def CheckFileStats(checkElement):
    success, msg = CheckFileExists(checkElement)
    if not success:
        return success, msg
    success, msg = ValidateElement(checkElement, "expect")
    if not success:
        return False, msg
    path = checkElement.get("path")
    expect = checkElement.get("expect")
    command = "stat -c '%U %G %a' " + path
    error,msg = waagent.RunGetOutput(command)
    msg = msg.strip()
    if ( msg == expect ):
        return True, 0
    return False, msg
    
# Return true if file permissions match or if file does not exist
def CheckFileStatsIfExists(checkElement):
    success, msg = CheckFileExists(checkElement)
    if not success:
        return True, 0
    return CheckFileStats(checkElement)

# Helper method for other functions
# Return True and the first matching line if a match is found
def GetMatchingLines(regex, path):
    if not (regex or path):
        return False, 0
    if not os.path.isfile(path):
        return False, 0
    with open(path) as f:
        for line in f:
            match = re.search(regex, line)
            if (match != None):
                return True, match
    return False, 0

# Return True if command runs successfully    
def RunCommand(command):
    if not command:
        return False, "Must supply a valid command to run"
    err,msg = waagent.RunGetOutput(command)
    if (err != 0):
        return False, "Error " + `err` + " running command: " + command
    lines = string.split(msg, "\n")
    return True, lines
    
# Return true if any of the outputted lines match the regex
# If false, return the entire output
def IsCommandOutputMatch(command, regex):
    if not regex:
        return False, "Must supply a valid regex"
    success, msg = RunCommand(command)
    if not success:
        return success, msg
    for line in msg:
        match = re.search(regex, line)
        if (match != None):
            return True, 0
    return False, string.join(msg, " ")
    
# Return true if none of the outputted lines match the regex
# If false, return only the matching line
def IsCommandOutputNotMatch(command, regex):
    if not regex:
        return False, "Must supply a valid regex"
    success, msg = RunCommand(command)
    if not success:
        return success, msg
    for line in msg:
        match = re.search(regex, line)
        if (match != None):
            return False, match.group(0)
    return True, 0  

# Return true if one or more matching lines are found
def CheckMatchingLines(checkElement):
    success, msg = CheckFileExists(checkElement)
    if not success:
        return False, msg
    success, msg = ValidateElement(checkElement, "regex")
    if not success:
        return False, msg
    path = checkElement.get("path")
    regex = checkElement.get("regex")
    success, msg = GetMatchingLines(regex, path)
    if success:
        return True, 0    
    return False, "Found no lines in " + path + " that match " + regex

# Return true if NO matching lines are found
def CheckNoMatchingLines(checkElement):
    success, msg = CheckFileExists(checkElement)
    if not success:
        return False, msg
    success, msg = ValidateElement(checkElement, "regex")
    if not success:
        return False, msg
    path = checkElement.get("path")
    regex = checkElement.get("regex")
    success, msg = GetMatchingLines(regex, path)
    if success:
        return False, "Found [" + msg.group(0) + "] in path " + path
    return True, 0
   
# Return true if one or more matching lines are found OR if file does not exist
def CheckMatchingLinesIfExists(checkElement):
    success, msg = CheckFileExists(checkElement)
    if not success:
        return True, msg
    return CheckMatchingLines(checkElement)

# Return true if no matching lines are found OR if file does not exist    
def CheckNoMatchingLinesIfExists(checkElement):
    success, msg = CheckFileExists(checkElement)
    if not success:
        return True, msg
    return CheckNotMatchingLines(checkElement)
    
# Return true if one or more matching lines are found in any files in the specified dir
def CheckMatchingLinesInDir(checkElement):
    success, msg = CheckFileExists(checkElement)
    if not success:
        return False, msg
    success, msg = ValidateElement(checkElement, "regex")
    if not success:
        return False, msg
    path = checkElement.get("path")
    regex = checkElement.get("regex")
    for filename in os.listdir(path):
        filePath=path + "/" + filename
        success, msg = GetMatchingLines(regex, filePath)
        if success:
            return True, 0
    return False, "Found no files matching " + regex + " in " + path
    
# Return true if one or more matching lines are found in ALL files in the specified dir
def CheckMatchingFilesInDir(checkElement):
    #TODO validate element
    success, msg = CheckFileExists(checkElement)
    if not success:
        return False, msg
    success, msg = ValidateElement(checkElement, "regex")
    if not success:
        return False, msg
    path = checkElement.get("path")
    regex = checkElement.get("regex")
    for filename in os.listdir(path):
        filePath=path + "/" + filename
        success, msg = GetMatchingLines(regex, filePath)
        if not success:
            return False, filePath + " does not match " + regex
    return True, 0
    
# Return true if no package updates are needed AND no error running command
def CheckAvailableUpdates(checkElement):
    success, result = patchscan.GetNeededPatches(util.distro)
    if not success:
        return False, "Error running command: " + result
    if (len(result) < 1):
        return True, "No patches needed"
    return False, `len(result)` + " Patches Needed"

# Return true if all packages are validated AND no error running command
def VerifyPackages(checkElement):
    command="dpkg -V "
    regex="^((?!/etc).)*$"
    return IsCommandOutputNotMatch(command, regex)
    
# Return true if kernel source is validated AND no error running command
def VerifyKernelSource(checkElement):
    success, msg = ValidateElement(checkElement, "regex")
    if not success:
        return False, msg
    regex = checkElement.get('regex')
    command = "uname -v"
    return IsCommandOutputMatch(command, regex)
    
# Return true if the environment variable for the specified user does not contain the specified text
def CheckNotMatchEnvVariable(checkElement):
    success, msg = ValidateElement(checkElement, "regex", "user", "variable")
    if not success:
        return False, msg
    regex = checkElement.get('regex')
    user = checkElement.get('user')
    variable = checkElement.get('variable')
    command="sudo -u " + user + " echo $" + variable
    return IsCommandOutputNotMatch(command, regex)

def CheckFirewallStatus(checkElement):
    success, msg = ValidateElement(checkElement, "regex")
    if not success:
        return False, msg
    regex = checkElement.get('regex')
    command = "ufw status verbose"
    return IsCommandOutputMatch(command, regex) 
    
def CheckServiceStatus(checkElement):
    success, msg = ValidateElement(checkElement, "regex", "service")
    if not success:
        return False, msg
    regex = checkElement.get('regex')
    service = checkElement.get('service')
    command = "service " + service + " status; :"
    return IsCommandOutputMatch(command, regex) 
    
# Return OK if the specified port is not explicitly allowed
def CheckFirewallBlock(checkElement):
    success, msg = ValidateElement(checkElement, "port", "protocol")
    if not success:
        return False, msg
    port = checkElement.get('port')
    protocol = checkElement.get('protocol')
    regex = "^ACCEPT"
    command = "iptables -nL | egrep '" + protocol + " [\S]+:" + port + "'; :"
    return IsCommandOutputNotMatch(command, regex)

# Return OK if no accounts have last login records older that are 70 days or older
def CheckStaleAccounts(checkElement):
    regex = "^.+$"
    command = "egrep -v ':/bin/false$|^root:x:0|^ubuntu:x:1000|^packer:x:1001' /etc/passwd | cut -d: -f 1 | while read user; do lastlog -b 70 -u $user | grep -v 'Latest$'; done; :"
    return IsCommandOutputNotMatch(command, regex)
    
# Return OK if no interfaces are configured in promiscuous mode
def CheckNoPromiscInterfaces(checkElement):
    command = "ifconfig"
    regex = "PROMISC"
    return IsCommandOutputNotMatch(command, regex)
    
# Return OK if no wireless interfaces are enabled
def CheckNoWirelessInterfaces(checkElement):
    command = "iwconfig 2>&1 | grep -v 'no wireless extensions'"
    regex = "^.+$"
    return IsCommandOutputNotMatch(command, regex)

def CheckWorldWriteableDirsWithoutStickyBit(checkElement):
    command = "find / -type d -perm -0002 ! -perm -1000 > /dev/null; :"
    regex = "^.+$"
    return IsCommandOutputNotMatch(command, regex)
    
def CheckSuidGUidPrograms(checkElement):
    ignoreElements = checkElement.findall('ignore')
    ignore = ""
    index = 1
    for ignoreElement in (ignoreElements):
        ignore+= "^" + ignoreElement.text
        if (index < len(ignoreElements)):
            ignore +="|"
        index+=1
    command = "find / -type f \( -perm -4000 -o -perm -2000 \) 2>/dev/null | egrep -v '" + ignore + "'; :"
    regex = "^.+$"
    return IsCommandOutputNotMatch(command, regex)
    
def CheckFileOwnership(checkElement):
    ignoreElements = checkElement.findall('ignore')
    ignore = ""
    index = 1
    for ignoreElement in (ignoreElements):
        ignore+= "^" + ignoreElement.text
        if (index < len(ignoreElements)):
            ignore +="|"
        index+=1
    command = "find / \( -nouser -o -nogroup \) 2>/dev/null | egrep -v '" + ignore + "'; :"
    regex = "^.+$"
    return IsCommandOutputNotMatch(command, regex)
    
def CheckWriteableDirsInPath(checkElement):
    success, msg = ValidateElement(checkElement, "user")
    if not success:
        return False, msg
    user = checkElement.get('user')
    command = "sudo -u " + user + " echo $PATH"
    success, msg = RunCommand(command)
    if not success:
        return success, msg
    line = msg[0]
    paths = string.split(line, ":")
    failPaths = ""
    for path in paths:
        if not (os.path.exists(path)):
            continue
        mode = os.stat(path).st_mode
        oWrite = bool(mode & stat.S_IWOTH)
        gWrite = bool(mode & stat.S_IWGRP)
        if (oWrite or gWrite):
            failPaths += " " + path
    if (len(failPaths) < 1):
        return True, 0
    return False, failPaths
    
def CheckHomeDirectoryPermissions(checkElement):
    command = "cat /etc/passwd | cut -d: -f6 | egrep -v '/bin|/dev|/usr|/var'"
    success, msg = RunCommand(command)
    if not success:
        return success, msg
    failPaths = ""
    for path in msg:
        if not (os.path.exists(path)):
            continue
        mode = os.stat(path).st_mode
        oRead = bool(mode & stat.S_IROTH)
        oWrite = bool(mode & stat.S_IWOTH)
        oExec = bool(mode & stat.S_IXOTH)
        gWrite = bool(mode & stat.S_IWGRP)
        if (oRead or oWrite or oExec or gWrite):
            failPaths += " " + path
    if (len(failPaths) < 1):
        return True, 0
    return False, failPaths
    
def CheckNetfs(checkElement):
    command = "mount -t nfs,nfs4,smbfs,cifs,ncpfs"
    regex = "^.+$"
    return IsCommandOutputNotMatch(command, regex)
    
# Return True if client signing is enabled OR file does not exist
def CheckSmbClientSigning(checkElement):
    filename = "/etc/samba/smb.conf"
    regex = "^\s*client\s+signing\s*=\s*mandatory"
    if not os.path.isfile(filename):
        return True, 0
    command = "awk '/^$/ || /^\s*[#;]/ {next} /^[[]/ {g=($1==\"[global]\");next} g {print}' " + filename
    return IsCommandOutputMatch(command, regex)
    
def CheckLibUser(checkElement):
    filename = "/etc/libuser.conf"
    regex1 = "[import]"
    regex2 = "^\s*login_defs\s*=\s*/etc/login.defs"
    regex3 = "[userdefaults]"
    regex4 = "^\s*LU_SHADOW(MAX|MIN|WARNING)\s*="
    if not os.path.isfile(filename):
        return True, 0
    success1, msg = GetMatchingLines(regex1, filename)
    success2, msg = GetMatchingLines(regex2, filename)
    if not (success1 and success2):
        return False, "login_defs not set in " + filename
    success3, msg = GetMatchingLines(regex3, filename)
    success4, msg = GetMatchingLines(regex4, filename)
    if success3 and not success4:
        return False, "LU_SHADOW(MAX|MIN|WARNING) found in " + filename
    return True, 0
    
# A given audit may have multiple elements representing different validation methods for different distros
# Find the method that works for this distro
# If the same element/check works for multipel distros, we can simply 
# concatenate them (e.g. distro="UbuntuDebian") in auditsXml
def FindCheckElement(auditElement, distro):
    checkElements = auditElement.findall("check")
    util.WriteDebug(inspect.stack()[0][3] + ": " + distro)
    for checkElement in checkElements:
        distroAttrib = checkElement.get("distro")
        if distro.lower() in  distroAttrib.lower():
            return (True, checkElement)
    return False, 0
    
# Call the appropriate function based on the specified command, or print an error message
def ProcessCheck(checkElement):
    # If command doesn't exist, result will be None, which will result in "command not supported"
    command = checkElement.get("command")
    util.WriteDebug(inspect.stack()[0][3] + ": " + command)
    if not command in globals().keys():
        return False, "Command not supported [" + command + "]"
    myFunction = globals()[command]
    return myFunction(checkElement)
    
# Return False if the sha256 sum of the audits file is does not match the expected value
# This is intended to prevent an attacker from manipulating the file to escalate privileges
def ValidateAuditsFile(filename):
    if not filename:
        util.WriteConsole(syslog.LOG_ERR, "Baseline scanning audit file must be a valid string")
        return False
    if not os.path.isfile(filename):
        util.WriteConsole(syslog.LOG_ERR, "No baseline scanning audit file could be found at path " + filename)
        return False
    sha = hashlib.sha256()
    f = open(filename)
    while True:
        data = f.read(128)
        if not data:
            break
        sha.update(data)
    f.close()
    digest = sha.hexdigest()
    util.WriteVerbose("Comparing hash sum of " + filename + ":")
    util.WriteVerbose("Expect: " + auditsHash)
    util.WriteVerbose("Actual: " + digest)
    if auditsHash != digest:
        util.WriteConsole(syslog.LOG_ERR, "Error validating file " + filename + ". Expected: " + auditsHash + "; Found: " + digest)
        return False
    return True
    
# Return true if file exists, hash is validated, and xml is parsable
# Also returns audits, which is a collection of xml audit elements
def ValidateAndParseAuditsFile(filename):
    if not ValidateAuditsFile(filename):
        return False, 0
    try:
        elementTree = parse(filename)
        root = elementTree.getroot()
        auditsRoot = root.find('audits')
        audits = auditsRoot.findall('audit')
    except:
        util.WriteConsole(syslog.LOG_ERR, "Unable to parse file as XML: " +  filename)
        return False, 0
    if (len(audits) < 1):
        return False, audits
    return True, audits
    
# Primary function for executing baseline scan. Returns a tuple containing:
# Exit code (0 = OK, 1 = err)
# Number of findings
def RunSecureBaselineScan(checkIndex, auditFile=auditsXml):   
    syslog.openlog(facility=util.defaultFacility)
    success, audits = ValidateAndParseAuditsFile(auditFile)
    if not success:
        util.WriteConsole(syslog.LOG_ERR, "Error validating baseline scanning audit file " + auditsXml + ". Scanning will NOT proceed")
        return 1, 0
    # If a specific audit was requested, find it and create a collection containing only that audit
    if (checkIndex):
        onlyAudit = (audit for audit in audits if audit.get('msid') == checkIndex)
        audits = []
        audits+=onlyAudit
        if len(audits) < 1:
            util.WriteConsole(syslog.LOG_ERR, "The specified audit [" + checkIndex + "] does not exist")
            return 1, 1
    util.WriteLog(syslog.LOG_INFO, "Processing " + `len(audits)` + " checks")
    failedAudits=0
    totalAudits=len(audits)
    for audit in audits:
        auditId = audit.get("cceid")
        description = audit.get("description")
        msId = audit.get("msid")
        util.WriteVerbose("Processing [" + msId + "] " + description)
        success, checkElement = FindCheckElement(audit, util.distro)
        if not success:
            util.WriteLog(syslog.LOG_WARNING, "Audit not supported: [" + msId + "] " + description)
            continue
        success, msg = ProcessCheck(checkElement)
        if not success:
            util.WriteLog(syslog.LOG_WARNING, "FAILED: [" + msId + "]" + description + " - Output:" + msg.strip())
            failedAudits+=1
    noun = "discrepancies"
    level = syslog.LOG_WARNING
    # Write summary
    if (failedAudits == 0):
        level=syslog.LOG_INFO
    elif (failedAudits == 1):
        noun = "discrepancy"
    util.WriteLog(level, "Baseline scan complete, found " + `failedAudits` + " " + noun + " out of " + `totalAudits` + " total audits")
    syslog.closelog()
    return 0, failedAudits